@@ -31,4 +31,7 @@ const char *ohcount_detect_language(SourceFile *sourcefile);
 
 int ohcount_is_binary_filename(const char *filename);
 
+/* Exported for unit testing */
+void escape_path(char *safe, const char *unsafe);
+
 #endif
