@@ -2,6 +2,11 @@
 #include <unistd.h>
 #include <errno.h>
 
+#ifdef _WIN32
+# include <fcntl.h>
+# define mkstemp(p) _open(_mktemp(p), _O_CREAT | _O_SHORT_LIVED | _O_EXCL)
+#endif
+
 /*
  * The bulk of this software is derived from Plan 9 and is thus distributed
  * under the Lucent Public License, Version 1.02, reproduced below.
@@ -619,9 +624,12 @@ static void output(int *added, int *removed) {
 char *tmp_file_from_buf(const char *buf) {
 	char *template = "/tmp/ohcount_diff_XXXXXXX";
 	char *path = strdup(template);
-	
+
 	int fd = mkstemp(path);
-	write(fd, buf, strlen(buf));
+  if (write(fd, buf, strlen(buf)) != strlen(buf)) {
+    fprintf(stderr, "src/diff.c: Could not write temporary file %s.\n", path);
+    exit(1);
+  }
 	close(fd);
 	return path;
 }
